/* <LIC_AMD_STD>
 * Copyright (c) 2004-2005 Advanced Micro Devices, Inc.
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 * The full GNU General Public License is included in this distribution in the
 * file called COPYING
 * </LIC_AMD_STD>  */
/* <CTL_AMD_STD>
 * </CTL_AMD_STD>  */
/* <DOC_AMD_STD>
 * Geode framebuffer driver
 * </DOC_AMD_STD>  */

#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/string.h>
#include <linux/fb.h>
#include <linux/init.h>
#include <linux/platform_device.h>

#include <asm/errno.h>

#include <cim/cim_parm.h>
#include <cim/cim_rtns.h>

/* module and KFB options */

char vmode[15] = {0};
int compression = 0;   /* Compression support default off */
int vfreq = 60;        /* vertical refresh default 60 */
int maxmem = (720 * 2 * 8192) / 1024;  /* 12288K */
int accel = 1;

#ifdef CONFIG_FB_GEODE_PANEL
int flatpanel = 1;  /* Flatpanel support default off when FP is enabled */
int fphwscale = 1;  /* HW scaling default on when FP is enabled*/
#endif

#ifdef MODULE

MODULE_AUTHOR("Advanced Micro Devices, Inc.");
MODULE_DESCRIPTION("Accelerated Framebuffer driver for AMD Geode devices");
MODULE_LICENSE("GPL");

MODULE_PARM(vfreq, "i");
MODULE_PARM_DESC(vfreq, "Set default video refresh rate");

MODULE_PARM(vmode, "s");
MODULE_PARM_DESC(vmode, "Set inital mode to <xres>x<vres>@refresh-bpp");

MODULE_PARM(compression, "i");
MODULE_PARM_DESC(compression, "Compression support (0=disabled or 1=enabled)");

MODULE_PARM(maxmem, "i");
MODULE_PARM_DESC(maxmem, "Maximum amount of memory to use for the FB (in K)");

MODULE_PARM(accel, "i");
MODULE_PARM_DESC(accel, "HW accleration support (0=disabled or 1=enabled)");

#ifdef CONFIG_FB_GEODE_PANEL
MODULE_PARM(flatpanel, "i");
MODULE_PARM_DESC(flatpanel, "Flatpanel support (0=disabled or 1=enabled)");

MODULE_PARM(fphwscale, "i");
MODULE_PARM_DESC(fphwscale, "FP Hardware Scaling (0=disabled or 1=enabled)");
#endif

#endif

#ifdef CONFIG_FB_GEODE_LX
extern int lxfb_init(struct device *);
extern void lxfb_cleanup(void);
extern int lxfb_suspend(struct device *dev, u32 state, u32 level);
extern int lxfb_resume(struct device *dev, u32 level);
#endif

struct geodefb_funcs {
  int cpu;
  int (*init)(struct device *);
  void (*cleanup)(void);
  int (*suspend)(struct device *, u32, u32);
  int (*resume)(struct device *, u32);
};

static struct geodefb_funcs geode_cpus[] = {
#ifdef CONFIG_FB_GEODE_LX
  { CIM_CPU_GEODELX, 
	  lxfb_init, 
	  lxfb_cleanup, 
	  lxfb_suspend, 
	  lxfb_resume },
#endif
  { 0, NULL, NULL }
};

static struct geodefb_funcs *geodefb_cpu = 0;

/* FIXME: __init? */

static int detect_geode_cpu(void) {
  unsigned long cpu, companion;
  int f = 0;

  /* This returns both the CPU and the companion chip */
  init_detect_cpu(&cpu, &companion);
  
  while(geode_cpus[f].cpu) {
    if (geode_cpus[f].cpu == (cpu & 0xFF)) {
      geodefb_cpu = &geode_cpus[f];
      break;
    }
    f++;
  }

  if (!geodefb_cpu) {
    printk(KERN_ERR "GEODEFB:  Unknown CPU detected (%lx)n", cpu);
    return -ENODEV;
  }
      
  return 0;
}

/* FIXME:  Don't build if we aren't a module? */

#define TYPE_INT    0x1
#define TYPE_STRING 0x2
#define TYPE_FLAG   0x3

struct geodefb_opts_t {
  char *option;
  int type;
  void *ptr;
  int size;
};

static struct geodefb_opts_t geodefb_options[] __initdata = {
  { "vfreq", TYPE_INT, &vfreq, sizeof(vfreq) },
  { "vmode", TYPE_STRING, vmode, sizeof(vmode) },
  { "compression", TYPE_FLAG, &compression, 0 },
  { "maxmem", TYPE_INT, &maxmem, sizeof(maxmem) },
  { "accel", TYPE_FLAG, &accel, 0 },
#ifdef CONFIG_FB_GEODE_PANEL
  { "flatpanel", TYPE_FLAG, &flatpanel, 0 },
  { "fphwscale", TYPE_FLAG, &fphwscale, 0 },
#endif
  { "", 0, 0 }
};

int __init
geodefb_setup(char *options) {

  char *opt;
  int o;

  if (!options || !*options)
    return 0;

  while((opt = strsep(&options, ","))) {
    if (!*opt) continue;
    
    for(o = 0; geodefb_options[o].type; o++) {
      struct geodefb_opts_t *p = &geodefb_options[o];

      if (p->type == TYPE_FLAG) {
	int flag = 1;
	char *ptr = opt;

	/* Allow them to say "no<token> */
	if (!strncmp(ptr, "no", 2)) {
	  flag = 0;
	  ptr += 2;
	}
				    
	if (!strcmp(ptr,p->option)) {
	  *((int *) p->ptr) = flag;
	  break;
	}

      } else {
	if (strncmp(opt, p->option, strlen(p->option))) 
	  continue;
	
	if (*(opt + strlen(p->option)) != '=')
	  continue;
	
	if (p->type == TYPE_INT) {
	  *((int *) p->ptr) = 
	    simple_strtoul(opt + strlen(p->option) + 1, NULL, 0);
	}
	else if (p->type == TYPE_STRING) {
	  strncpy((char *) p->ptr, opt + strlen(p->option) + 1, p->size);
	}
	
	break;
      }
    }
  }

  return 0;
}

static int geodefb_suspend(struct device *dev, u32 state, u32 level) {
  if (geodefb_cpu)
    return geodefb_cpu->suspend(dev, state, level);
  return 0;
}

static int geodefb_resume(struct device *dev, u32 level) {
  if (geodefb_cpu)
    return geodefb_cpu->resume(dev, level);
  return 0;
}

static int __init 
geodefb_probe(struct device *device) {
  
  if (detect_geode_cpu()) return -ENODEV;
  return geodefb_cpu->init(device);
}

static struct device_driver geodefb_driver = {
        .name   = "geodefb",
        .bus    = &platform_bus_type,
        .probe  = geodefb_probe,
	.suspend = geodefb_suspend,
	.resume = geodefb_resume
};

static struct platform_device geodefb_device = {
        .name   = "geodefb",
};


int __init
geodefb_init(void)
{
  char *option = NULL;
  int ret;

  if (fb_get_options("geodefb", &option))
    return -ENODEV;
  
  geodefb_setup(option);
  
  ret = driver_register(&geodefb_driver);

  if (!ret) {    
    ret = platform_device_register(&geodefb_device);

    if (ret)
      driver_unregister(&geodefb_driver);
  }

  return ret;
}

#ifdef MODULE
void __exit
geodefb_exit(void)
{
  if (geodefb_cpu)
    geodefb_cpu->cleanup();
}
#endif

module_init(geodefb_init);

#ifdef MODULE
module_exit(geodefb_exit);
#endif
